import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
import numpy as np
import copy

def crt_net(hs,dropout=0.0,activation=0,bias=True,bn=0):
    net = []
    for h0,h1 in zip(hs, hs[1:]):
        if bn == 1:
            net.extend([nn.Linear(h0, h1,bias=bias),nn.BatchNorm1d(h1),nn.ReLU(),nn.Dropout(p=dropout)])
        else:
            net.extend([nn.Linear(h0, h1,bias=bias),nn.ReLU(),nn.Dropout(p=dropout)])            
    n = 2
    if bn == 1:
        n = 3
    for i in range(n):
        net.pop()
    return nn.Sequential(*net)

def conv3x3proto(in_channels, out_channels, **kwargs):
    return nn.Sequential(
        nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, **kwargs),
        nn.BatchNorm2d(out_channels),
        nn.ReLU(),
        nn.MaxPool2d(2)
    )

class MetaGMM(nn.Module):
    def __init__(self,n_att,n_hiddens=[256,2],dropout=0.1,use_encoder=1,em_epoch=10,alpha0=1e-3,K=10,eps=1e-4,eps0=1e-8,use_ari=1,endeps=1e-8,temperature=1,cal_r=0,image=0,hidden_size=32,n_class=None):
        super(MetaGMM, self).__init__()
        if image >= 1:
            in_channels = 1
            last_hidden_size = hidden_size*36
            if image == 2:
                in_channels = 3
                last_hidden_size = hidden_size*25
            self.f_conv = nn.Sequential(
                conv3x3proto(in_channels, hidden_size),
                conv3x3proto(hidden_size, hidden_size),
                conv3x3proto(hidden_size, hidden_size),
                conv3x3proto(hidden_size, hidden_size)
            )
            hs = [last_hidden_size]+n_hiddens
            self.f_z = crt_net(hs,dropout=dropout)
            hs = list(reversed(n_hiddens))+[n_att]
        else:
            if use_encoder == 1:
                hs = [n_att]+n_hiddens
                self.f_z = crt_net(hs,dropout=dropout)
                hs = list(reversed(n_hiddens))+[n_att]

        if cal_r == 1:
            if use_encoder == 1:
                hs = [n_hiddens[-1]]+n_hiddens
            else:
                hs = [n_att]+n_hiddens
            self.f_t = crt_net(hs,dropout=dropout)
            hs = [n_hiddens[-1]]+n_hiddens
            self.g_t = crt_net(hs,dropout=dropout)
            if use_encoder == 1:
                hs = [n_hiddens[-1]*2]+n_hiddens[:-1]+[K]
            else:
                hs = [n_att+n_hiddens[-1]]+n_hiddens[:-1]+[K]
            self.f_r = crt_net(hs,dropout=dropout)

        self.log_beta0 = nn.Parameter(torch.zeros((1)))
        self.log_nu0_D = nn.Parameter(torch.zeros((1)))
        self.W0 = torch.eye(n_hiddens[-1])
        self.alpha0 = alpha0
        self.use_encoder = use_encoder
        self.em_epoch = em_epoch
        self.K = K
        if use_encoder == 1:
            self.D = n_hiddens[-1]
        else:
            self.D = n_att
        self.eps = eps
        self.eps0 = eps0
        self.endeps = endeps
        self.use_ari = use_ari
        self.temperature = temperature
        self.cal_r = cal_r
        self.image = image
        self.hidden_size = hidden_size
            
    def vb_spherical(self,z,r):
        ###input
        #z: (hidden) attribute vectors
        #   [number of samples] x [number of (hidden) attributes]
        #r: estimated cluster assignment probability
        #   [number of clusters] x [number of samples]
        #
        ###output
        #rs: list of estimated cluster assignment probabilities for each EM epoch
        #    list of [number of clusters] x [number of samples]
        
        n_sample,n_dim = z.shape
        prer = 0.0
        ak = torch.ones(self.K).to(z.device)
        bk = torch.ones(self.K).to(z.device)
        rs = []
        for i in range(self.em_epoch):
            #M-step
            gammak1 = 1.0+r.sum(1)
            gammak2 = self.alpha0+r.sum(1,keepdim=True).repeat(1,self.K).transpose(0,1).triu(diagonal=1).sum(1)
            muk = (bk/ak).unsqueeze(1)*(r.unsqueeze(2)*z.unsqueeze(0)).sum(1)/(1+(r.sum(1)*bk/ak).unsqueeze(1)) #KxD
            ak = 1+0.5*self.D*r.sum(1)
            bk = 1+0.5*(r*((z.unsqueeze(0)-muk.unsqueeze(1)).pow(2).sum(2)+self.D)).sum(1)

            #E-step
            digamma12 = torch.digamma(gammak1+gammak2)
            log_pi = torch.digamma(gammak1)-digamma12
            log_pi2 = torch.digamma(gammak2)-digamma12
            dig = log_pi2.unsqueeze(1).repeat(1,self.K).transpose(0,1).tril(diagonal=-1).sum(1)
            zm = z.unsqueeze(0)-muk.unsqueeze(1) #KxNxD            
            q = 0.5*self.D*(torch.digamma(ak)-bk.log()).unsqueeze(1)-0.5*(ak/bk).unsqueeze(1)*(zm.pow(2).sum(2)+self.D)
            log_r = (log_pi+dig).unsqueeze(1)+q
            r = torch.softmax(log_r/self.temperature,axis=0)
            rs.append(r)
            e = (r-prer).abs().sum()
            if i>0 and e < self.endeps:
                break
            prer = r
        return rs

    def encode(self,x):
        n_sample = x.shape[0]
        if self.image == 1 and self.use_encoder == 1: 
            x = x.view(n_sample,1,105,105)
            x = self.f_conv(x)
            x = x.view(n_sample,-1)
            x = self.f_z(x)            
        elif self.image == 2 and self.use_encoder == 1:
            x = x.view(n_sample,3,84,84)
            x = self.f_conv(x)
            x = x.view(n_sample,-1)
            x = self.f_z(x)            
        elif self.use_encoder == 1:
            x = self.f_z(x)
        return x
        
    def forward(self,x,y):
        n_sample = x.shape[0]
        x = self.encode(x)

        rs = []
        if self.cal_r == 1:
            t = self.g_t(self.f_t(x).sum(0)) #task vector
            xt = torch.cat((x,t.unsqueeze(0).repeat(n_sample,1)),axis=1)
            r = self.f_r(xt).transpose(0,1)/self.temperature #responsibility
            rs.append(r)
        else:
            eps = 100
            r = eps*torch.randn(self.K,n_sample).to(x.device)
        r = torch.softmax(r,axis=0)
        rs0 = self.vb_spherical(x,r)
        rs.extend(rs0)
        
        rmax = F.one_hot(torch.argmax(rs[-1],axis=0),num_classes=self.K).transpose(1,0)
        rmax_ari = self.calari(rmax,y,ari=self.use_ari)
        rlast_ari = self.calari(rs[-1],y,ari=self.use_ari)
        return -rlast_ari, rmax_ari, x, rs[-1]

    def calari(self,r,y,ari=1):
        ###input
        #r: estimated cluster assignment probability
        #   [number of clusters] x [number of samples]
        #y: one-hot true categories:
        #   [number of categories] x [number of samples]
        #ari: 0: Rand index
        #     1: adjusted Rand index (total variation distance-based)
        #     2: normalized mutual information
        #     3: adjusted Rand index (probability-based)
        ###output
        #ari 
        
        if ari == 2:
            eps = 1e-5
            r = r[r.sum(1)>0,:]
            n = (r.unsqueeze(0)*y.unsqueeze(1)).sum(2)
            N = float(r.shape[1])
            a = n.sum(axis=0,keepdim=True)
            b = n.sum(axis=1,keepdim=True)
            I2 = (N*n/(a*b)+eps).log()
            I = (n[n>0]/N*I2[n>0]).sum()
            Hu = -(a/N*(a/N+eps).log()).sum()
            Hv = -(b/N*(b/N+eps).log()).sum()
            NMI = 2*I/(Hu+Hv)
            return NMI
        else:
            if ari == 3: #proabbility
                dist_r = 1-(r.unsqueeze(1)*r.unsqueeze(2)).sum(0)
            else: #total variation distance
                dist_r = 0.5*(r.unsqueeze(1)-r.unsqueeze(2)).abs().sum(0)
            dist_y = 0.5*(y.unsqueeze(1)-y.unsqueeze(2)).abs().sum(0)
            a = ((1-dist_y)*(1-dist_r)).triu(diagonal=1).sum()
            b = ((1-dist_y)*dist_r).triu(diagonal=1).sum()
            c = (dist_y*(1-dist_r)).triu(diagonal=1).sum()
            d = (dist_y*dist_r).triu(diagonal=1).sum()
            if ari == 1 or ari == 3:
                ARI = 2*(a*d-b*c)/((a+b)*(b+d)+(a+c)*(c+d))
                return ARI
            else:
                RI = (a+d)/(a+b+c+d)
                return RI
        
    def reparameterize(self,mu,logvar):
        std = torch.exp(0.5*logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add_(mu)
    
    def proto(self,x,y,n_spt=10):
        y = y.t()
        x = self.encode(x)
        n_sample = x.shape[0]
        n_spt = int(n_sample*0.5)
        perm = np.random.permutation(n_sample)
        x_spt = x[perm[:n_spt],:]
        x_qry = x[perm[n_spt:],:]
        y_spt = y[perm[:n_spt],:]
        y_qry = y[perm[n_spt:],:]
        m = (x_spt.unsqueeze(1)*y_spt.unsqueeze(2)).sum(0)/(y_spt.sum(0).unsqueeze(1)+1e-5)
        dist = (m.unsqueeze(0)-x_qry.unsqueeze(1)).pow(2).sum(2)
        log_p = F.log_softmax(-dist,dim=1)
        loss = -log_p[y_qry==1].sum()/y_qry.shape[0]
        return loss
